import pytest

from . import create_app
from . import assert_export


@pytest.fixture
def app():
    return create_app()


@pytest.fixture
def client(app):
    _client = app.test_client()
    return _client


def test_app_runs(client):
    res = client.get('/')
    assert res.status_code == 404


def test_resources_list_json(client):
    res = client.get('/resources')
    assert res.status_code == 200
    assert res.headers.get('Content-Type') == 'application/json'


def test_resources_list_xml(client):
    res = client.get(
        '/resources',
        headers={'Accept': 'application/xml'}
    )
    assert res.status_code == 200
    assert res.headers.get('Content-Type') == 'application/xml; charset=utf-8'


def test_bad_request(client):
    res = client.post('/artist')
    assert res.status_code == 400

    res = client.put('/artist/1')
    assert res.status_code == 400


def test_not_found(client):
    res = client.delete('/artist/1000000000')
    assert res.status_code == 404

    res = client.get('/artist/1000000000')
    assert res.status_code == 404


def test_conflict(client):
    res = client.post(
        '/artist',
        json={'Name': 'Accept'},
        headers={'Accept': 'application/xml'}
    )
    assert res.status_code == 409
    assert res.headers.get('Content-Type') == 'application/xml; charset=utf-8'


def test_update(client):
    res = client.get('/artist/1')
    etag = res.headers.get('ETag')

    res = client.put(
        '/artist/1',
        json={'Name': 'pippo2'},
        headers={'If-Match': etag}
    )
    etag = res.headers.get('ETag')
    assert res.status_code == 200
    assert etag is not None

    res = client.patch(
        '/artist/1',
        json={'Name': 'pippo3'},
        headers={'If-Match': etag}
    )
    assert res.status_code == 200


def test_unprocessable_entity(client):
    res = client.post(
        '/artist',
        json={'pippo': 'pluto'}
    )
    assert res.status_code == 422
    assert res.headers.get('Content-Type') == 'application/json'

    res = client.put(
        '/artist/1',
        json={'pippo': 'pluto'}
    )
    assert res.status_code == 422
    assert res.headers.get('Content-Type') == 'application/json'

    res = client.patch(
        '/artist/1',
        json={'pippo': 'pluto'}
    )
    assert res.status_code == 422
    assert res.headers.get('Content-Type') == 'application/json'


def test_resource_crud(client):
    res = client.post(
        '/artist',
        json={'Name': 'pippo'}
    )
    assert res.status_code == 201
    assert res.headers.get('Content-Type') == 'application/json'

    data = res.get_json()
    id = data.get('ArtistId')

    etag = res.headers.get('ETag')
    assert etag is not None
    assert res.headers.get('Location').endswith('/artist/{}'.format(id))

    res = client.get('/artist/{}'.format(id))
    etag = res.headers.get('ETag')

    assert res.status_code == 200
    assert etag is not None
    assert res.headers.get('Content-Type') == 'application/json'
    assert res.headers.get('Link') == "</artist/{id}>; rel=self, </artist/{id}/album>; rel=related".format(id=id)

    data = res.get_json()
    returned_id = data.get('ArtistId')
    assert returned_id == id

    res = client.get('/artist/{}'.format(id), headers={'If-None-Match': etag})
    assert res.status_code == 304

    res = client.get('/artist/{}'.format(id), headers={'If-None-Match': 'fake_etag'})
    assert res.status_code == 200

    etag = res.headers.get('ETag')
    res = client.delete('/artist/{}'.format(id))
    assert res.status_code == 428

    res = client.delete('/artist/{}'.format(id), headers={'If-Match': 'fake_etag'})
    assert res.status_code == 412

    res = client.delete('/artist/{}'.format(id), headers={'If-Match': etag})
    assert res.status_code == 204


def test_put_creation(client):
    id_res = 123456

    res = client.put(
        '/some_model/{}'.format(id_res),
        json={'value': 'pippo'}
    )
    assert res.status_code == 201
    etag = res.headers.get('ETag')
    assert etag is not None
    assert res.get_json().get('id') == id_res

    res = client.get('/some_model/{}'.format(id_res), headers={'If-Match': etag})
    assert res.status_code == 200
    etag = res.headers.get('ETag')
    assert etag is not None
    assert res.get_json().get('id') == id_res

    res = client.delete('/some_model/{}'.format(id_res), headers={'If-Match': etag})
    assert res.status_code == 204


def test_put_failed(client):
    res = client.put(
        '/some_model/10',
        json={'pippo': 'pluto'}
    )
    assert res.status_code == 422

    data = res.get_json()['response']
    missing = data.get('missing')
    unknown = data.get('unknown')
    assert len(missing) == 1
    assert len(unknown) == 1
    assert missing[0] == 'value'
    assert unknown[0] == 'pippo'


def test_resource_meta(client):
    res = client.get('/artist/meta')
    assert res.status_code == 200
    assert res.headers.get('Content-Type') == 'application/json'

    data = res.get_json()
    assert all(i in data.keys() for i in (
        'name',
        'description',
        'fields',
        'methods',
        'url'
    ))


def test_get_list(client):
    res = client.get('/artist')
    assert res.status_code == 200


def test_hateoas(client):
    res = client.get('/artist/1')
    assert res.status_code == 200

    data = res.get_json()
    assert '_links' in data.keys()
    assert data['_links'].get('self') == '/artist/1'


def test_export(client):
    res = client.get('/track?_export=pippo')
    assert res.status_code == 200
    assert_export(res, 'pippo')


def test_subresource(client):
    res = client.get('/album/5/track?_extended')
    assert res.status_code == 200

    data = res.get_json()['TrackList'][0]
    assert data['AlbumId'] == 5
    assert all(e in data.keys() for e in (
        "TrackId",
        "GenreId",
        "MediaTypeId",
        "Album",
        "Genre",
        "MediaType"
    ))



from flask import Flask
from flask.testing import FlaskClient

from flask_autocrud import AutoCrud

from .models import db


def create_app(m=None, conf=None):
    class TestClient(FlaskClient):
        def fetch(self, url, *args, **kwargs):
            return self.open(url, method='FETCH', *args, **kwargs)

    _app = Flask(__name__)
    _app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite+pysqlite:///tests/db.sqlite3'
    _app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
    _app.config.update(**(conf or {}))

    with _app.app_context():
        db.init_app(_app)
        AutoCrud(_app, db, models=m)

    _app.test_client_class = TestClient
    _app.testing = True
    return _app


def assert_pagination(res, code, page, limit):
    assert res.status_code == code
    assert res.headers.get('Pagination-Page') == page
    assert res.headers.get('Pagination-Page-Size') == limit
    assert res.headers.get('Pagination-Count') is not None
    assert res.headers.get('Pagination-Num-Pages') is not None
    assert res.headers.get('Link') != ''

    data = res.get_json()
    assert '_meta' in data
    assert all(e in data['_meta'].keys() for e in (
        "first",
        "last",
        "next",
        "prev"
    ))


def assert_export(res, filename):
    assert res.headers.get('Total-Rows') is not None
    assert res.headers.get('Total-Columns') is not None
    assert res.headers.get('Content-Type') == 'text/csv; charset=utf-8'
    assert res.headers.get('Content-Disposition') == 'attachment; filename={}.csv'.format(filename)



from sqlalchemy.orm.mapper import Mapper
from sqlalchemy.inspection import inspect
from sqlalchemy.orm.properties import ColumnProperty

from .config import ALLOWED_METHODS


class Model(object):
    __pks__ = None
    __url__ = None
    __cols__ = None
    __rels__ = None
    __table__ = None
    __hidden__ = []
    __version__ = '1'
    __description__ = None
    __methods__ = ALLOWED_METHODS

    collection_suffix = 'List'

    def __str__(self):
        """

        :return: (str) primary key value
        """
        return str(getattr(self, self.primary_key_field()))

    @classmethod
    def _load_cols(cls):
        """

        """
        cls.__pks__ = []
        cls.__cols__ = {}

        for i in cls.__dict__:
            if not i.startswith('_') and i not in cls.__hidden__:
                col = getattr(cls, i)
                if isinstance(col.comparator, ColumnProperty.Comparator):
                    cls.__cols__[i] = col
                    if col.primary_key:
                        cls.__pks__.append(i)

    @classmethod
    def _load_related(cls):
        """

        """
        cls.__rels__ = {}

        for r in inspect(cls).relationships:
            if isinstance(r.argument, Mapper):
                key = r.argument.class_.__name__
                columns = r.argument.class_().columns()
            else:
                key = r.argument.__name__
                columns = r.argument.columns()

            instance = getattr(cls, r.key)
            cls.__rels__.update({key: dict(instance=instance, columns=columns)})

    @classmethod
    def columns(cls):
        """

        :return:
        """
        if cls.__cols__ is None:
            cls._load_cols()
        return cls.__cols__

    @classmethod
    def required(cls):
        """

        :return:
        """
        columns = []
        for col, c in cls.columns().items():
            if not (c.nullable or c.primary_key) or (c.primary_key and not c.autoincrement):
                columns.append(col)
        return columns

    @classmethod
    def searchable(cls):
        """

        :return:
        """
        columns = []
        for col, c in cls.columns().items():
            if c.type.python_type is str:
                columns.append(col)
        return columns

    @classmethod
    def optional(cls):
        """

        :return:
        """
        columns = []
        for col, c in cls.columns().items():
            if c.nullable:
                columns.append(col)
        return columns

    @classmethod
    def primary_key_field(cls):
        """

        :return:
        """
        if cls.__pks__ is None:
            cls._load_cols()
        return cls.__pks__[0]

    @classmethod
    def related(cls, name=None):
        """

        :param name:
        :return:
        """
        if not cls.__rels__:
            cls._load_related()

        if name is None:
            return cls.__rels__

        for k in cls.__rels__.keys():
            if k == name:
                return cls.__rels__.get(k).get('instance'),\
                       cls.__rels__.get(k).get('columns')
        return None, None

    @classmethod
    def submodel_from_url(cls, url):
        """

        :param url:
        :return:
        """
        for r in inspect(cls).relationships:
            if isinstance(r.argument, Mapper):
                if r.argument.class_.__url__ == url:
                    return r.argument.class_

    @classmethod
    def validate(cls, data):
        """

        :param data:
        """
        fields = cls.required() + cls.optional()
        unknown = [k for k in data if k not in fields]
        missing = list(set(cls.required()) - set(data.keys()))

        return missing if len(missing) else None, unknown if len(unknown) else None

    @classmethod
    def description(cls):
        """

        :return:
        """
        return dict(
            url=cls.__url__,
            name=cls.__name__,
            methods=list(cls.__methods__),
            description=cls.__description__ or cls.__table__.comment,
            fields=[
                dict(
                    name=col,
                    type=c.type.python_type.__name__,
                    key=c.primary_key,
                    autoincrement=c.autoincrement,
                    nullable=c.nullable,
                    unique=c.unique,
                    description=c.comment
                ) for col, c in cls.columns().items()
            ]
        )

    def to_dict(self, links=False):
        """

        :param links:
        :return:
        """
        resp = {}
        data = self if isinstance(self, dict) else self.__dict__

        for k, v in data.items():
            if k.startswith('_'):
                continue

            if isinstance(v, Model):
                resp.update({v.__class__.__name__: v.to_dict(links)})
            elif isinstance(v, list):
                if len(v) > 0:
                    name = v[0].__class__.__name__ + self.collection_suffix
                    resp.update({name: [i.to_dict(links) for i in v]})
            else:
                resp.update({k: v})

        if links:
            resp['_links'] = self.links()
        return resp

    def links(self):
        """

        :return:
        """
        link_dict = dict(self=self.resource_uri())

        for r in inspect(self.__class__).relationships:
            if not r.uselist:
                instance = getattr(self, r.key)
                if instance:
                    key = r.argument.__name__
                    link_dict[key] = instance.resource_uri()
            elif isinstance(r.argument, Mapper):
                key = r.argument.class_.__name__
                link_dict[key] = "{}{}".format(self.resource_uri(), r.argument.class_.__url__)

        return link_dict

    def resource_uri(self):
        """

        :return:
        """
        pk = getattr(self, self.primary_key_field())
        return "{}/{}".format(self.__url__, pk) if pk else None

    def update(self, attributes):
        """

        :param attributes:
        :return:
        """
        for attr, val in attributes.items():
            if attr in self.columns().keys():
                setattr(self, attr, val)
        return self



from colander import null
from colander import String
from colander import Invalid
from colander import SchemaType
from colander import SchemaNode
from colander import MappingSchema
from colander import SequenceSchema


class FilterValue(SchemaType):
    def deserialize(self, node, cstruct):
        if cstruct is null:
            return null
        t = type(cstruct)

        if t not in (str, int, float, list):
            raise Invalid(node, "Invalid types, only str, int, float, list are supported")

        return cstruct


class RelatedSchema(SchemaType):
    def deserialize(self, node, cstruct):
        if cstruct is null:
            return null

        if type(cstruct) is not dict:
            raise Invalid(node, "Invalid types: it must be dict")

        for k, v in cstruct.items():
            if type(v) is not list:
                raise Invalid(node, "Invalid types of '{}': it must be a list".format(k))
            for i in v:
                if type(i) is not str:
                    raise Invalid(node, "Invalid types of '{}[{}]': it must be a str".format(k, v.index(i)))

        return cstruct


class FilterSchema(MappingSchema):
    model = SchemaNode(String())
    field = SchemaNode(String())
    op = SchemaNode(String())
    value = SchemaNode(FilterValue())


class SortingSchema(MappingSchema):
    model = SchemaNode(String())
    field = SchemaNode(String())
    direction = SchemaNode(String())


class Filters(SequenceSchema):
    filter = FilterSchema()


class Sorting(SequenceSchema):
    sort = SortingSchema()


class FieldsSchema(SequenceSchema):
    fields = SchemaNode(String())


class FetchPayloadSchema(MappingSchema):
    filters = Filters(missing=[])
    sorting = Sorting(missing=[])
    fields = FieldsSchema(missing=[])
    related = SchemaNode(RelatedSchema(), missing={})



# noinspection PyUnresolvedReferences
from waitress import serve


class StandaloneApplication:
    def __init__(self, app, options=None):
        """

        :param app:
        :param options:
        """
        self._app = app
        self._options = options or {}

    def run(self):
        """

        """
        serve(
            self._app,
            listen=self._options.get('bind'),
            threads=self._options.get('workers')
        )



__version_info__ = (2, 1, 1)
__version__ = '.'.join(map(str, __version_info__))

__author_info__ = {
    'name': 'cs91chris',
    'email': 'cs91chris@voidbrain.me'
}

__author__ = "{} <{}>".format(
    __author_info__['name'],
    __author_info__['email']
)

__all__ = [
    '__version__',
    '__version_info__',
    '__author__',
    '__author_info__'
]



from flask import Blueprint

from sqlalchemy.ext.automap import automap_base
from sqlalchemy.ext.declarative import declarative_base

from flask_errors_handler import ErrorHandler
from flask_response_builder import ResponseBuilder

from .model import Model
from .service import Service
from .config import set_default_config


class AutoCrud(object):
    def __init__(self, app=None, db=None, models=None, builder=None, error=None, **kwargs):
        """

        :param db:
        :param app:
        :param models:
        :param builder:
        :param error:
        """
        self._app = app
        self._db = db
        self._api = None
        self._models = {}
        self._response_error = None
        self._response_builder = None

        if app is not None:
            self.init_app(
                self._app, self._db,
                models=models, builder=builder, error=error,
                **kwargs
            )

    @property
    def blueprint(self):
        """

        :return:
        """
        return self._api

    @property
    def response_builder(self):
        """

        :return:
        """
        return self._response_builder

    @property
    def error_handler(self):
        """

        :return:
        """
        return self._response_error

    @property
    def models(self):
        """

        :return:
        """
        return self._models

    def init_app(self, app, db, models=None, builder=None, error=None, **kwargs):
        """

        :param app:
        :param db:
        :param models:
        :param builder:
        :param error:
        :return:
        """
        self._db = db
        self._app = app
        self._response_error = error or ErrorHandler()
        self._response_builder = builder or ResponseBuilder()

        if not isinstance(self._response_builder, ResponseBuilder):
            raise AttributeError("'builder' must be instance of '{}'".format(ResponseBuilder))

        if not isinstance(self._response_error, ErrorHandler):
            raise AttributeError("'error' must be instance of '{}'".format(ErrorHandler))

        set_default_config(self._app)
        self._response_builder.init_app(self._app)
        self._response_error.init_app(self._app, response=self._response_builder.on_accept())

        subdomain = self._app.config['AUTOCRUD_SUBDOMAIN']
        self._api = Blueprint('flask_autocrud', __name__, subdomain=subdomain)

        if models is not None:
            for m in models:
                if not issubclass(m, (db.Model, Model)):
                    raise AttributeError(
                        "'{}' must be both a subclass of {} and of {}".format(m, db.Model, Model)
                    )
                self._register_model(m, **kwargs)
        else:
            schema = self._app.config['AUTOCRUD_DATABASE_SCHEMA']
            automap_model = automap_base(declarative_base(cls=(db.Model, Model)))
            automap_model.prepare(self._db.engine, reflect=True, schema=schema)

            for model in automap_model.classes:
                self._register_model(model, **kwargs)
                if self._app.config['AUTOCRUD_READ_ONLY']:
                    model.__methods__ = {'OPTION', 'HEAD', 'GET', 'FETCH'}

                if not self._app.config['AUTOCRUD_FETCH_ENABLED']:
                    model.__methods__ -= {'FETCH'}

        if self._app.config['AUTOCRUD_RESOURCES_URL_ENABLED']:
            self._register_resources_route()

        self._response_error.api_register(self._api)
        self._app.register_blueprint(self._api)

        if not hasattr(app, 'extensions'):
            app.extensions = dict()
        app.extensions['autocrud'] = self

    def _register_model(self, model, **kwargs):
        """

        :param model:
        """
        class_name = model.__name__
        if model.__url__ is None:
            model.__url__ = "{}/{}".format(
                self._app.config['AUTOCRUD_BASE_URL'],
                class_name.lower()
            )

        view = type(
            class_name + 'Service',
            (Service,), {
                '_model': model,
                '_db': self._db,
                '_response': self._response_builder,
                **kwargs
            }
        ).as_view(class_name)

        def add_route(url='', methods=None, **kwargs):
            self._api.add_url_rule(
                model.__url__ + url,
                view_func=view,
                methods=methods,
                strict_slashes=False,
                **kwargs
            )

        add_route(methods=['POST', 'FETCH'])
        add_route(defaults={'resource_id': None})

        if self._app.config['AUTOCRUD_METADATA_ENABLED'] is True:
            add_route(self._app.config['AUTOCRUD_METADATA_URL'])

        self._models[model.__name__] = model
        pk = model.columns().get(model.primary_key_field())
        pk_type = pk.type.python_type.__name__
        add_route('/<{}:{}>'.format(pk_type, 'resource_id'), model.__methods__ - {'POST', 'FETCH'})
        add_route('/<{}:{}>/<path:subresource>'.format(pk_type, 'resource_id'), {'GET'})

    def _register_resources_route(self):
        """

        :return:
        """
        url = self._app.config['AUTOCRUD_BASE_URL'] + self._app.config['AUTOCRUD_RESOURCES_URL']

        @self._api.route(url)
        @self.response_builder.on_accept()
        def index():
            return {
                res: "{}{{/{}}}".format(cls.__url__, cls.primary_key_field())
                for res, cls in self._models.items()
            }



from flask import Flask

from flask_admin import Admin
from flask_admin.contrib.sqla import ModelView
from sqlalchemy.orm import relationship

from flask_autocrud import Model
from flask_autocrud import AutoCrud
from flask_sqlalchemy import SQLAlchemy

from sqlalchemy import ForeignKey

db = SQLAlchemy()


class CustomAdminView(ModelView):
    column_display_pk = True
    can_set_page_size = True
    can_view_details = True
    details_modal = True


class artists(db.Model, Model):
    __tablename__ = "Artist"
    __description__ = 'artists'
    id = db.Column('ArtistId', db.Integer, primary_key=True, comment='primarykey')
    name = db.Column('Name', db.String(80), unique=True, nullable=False)


class albums(db.Model, Model):
    __tablename__ = "Album"
    __description__ = 'my albums table'
    id = db.Column('AlbumId', db.Integer, primary_key=True)
    title = db.Column('Title', db.String(80), unique=True, nullable=False, comment="column description")
    artist_id = db.Column('ArtistId', db.Integer, ForeignKey("Artist.ArtistId"), nullable=False)
    # NOTE: it must be the same name of related class otherwise there will be unknown side effects
    # This will change in future versions
    artists = relationship(artists, backref="albums")


def main():
    app = Flask(__name__)
    app.config['SECRET_KEY'] = 'more_difficult_string'
    app.config['RB_DEFAULT_ACCEPTABLE_MIMETYPES'] = ['application/json', 'application/xml']
    app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite+pysqlite:///examples/db.sqlite3'
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False

    db.init_app(app)
    admin = Admin(app)
    autocrud = AutoCrud(app, db, models=[artists, albums])

    for k, m in autocrud.models.items():
        setattr(CustomAdminView, 'column_searchable_list', m.searchable())
        admin.add_view(CustomAdminView(m, db.session))

    app.run(debug=True)


if __name__ == '__main__':
    main()



class StandaloneApplication:
    def __init__(self, app, *args, **kwargs):
        """

        :param app:
        :param args:
        :param kwargs:
        """
        self._app = app

    def run(self):
        """

        """
        self._app.run()



import pytest

from . import create_app


@pytest.fixture
def app():
    return create_app(conf={
        'AUTOCRUD_READ_ONLY': True,
        'AUTOCRUD_FETCH_ENABLED': False,
        'AUTOCRUD_EXPORT_ENABLED': False,
        'AUTOCRUD_METADATA_ENABLED': False,
        'AUTOCRUD_RESOURCES_URL_ENABLED': False,
        'AUTOCRUD_CONDITIONAL_REQUEST_ENABLED': False,
    })


@pytest.fixture
def client(app):
    _client = app.test_client()
    return _client


def test_read_only(client):
    res = client.post(
        '/artist',
        json={'Name': 'Accept'}
    )
    assert res.status_code == 405

    res = client.put(
        '/artist/1',
        json={'Name': 'pippo2'}
    )
    assert res.status_code == 405

    res = client.patch(
        '/artist/1',
        json={'Name': 'pippo3'}
    )
    assert res.status_code == 405

    res = client.delete('/artist/1')
    assert res.status_code == 405


def test_get_resource(client):
    res = client.get('/artist/1')
    assert res.status_code == 200

    etag = res.headers.get('ETag')
    assert etag is None

    data = res.get_json()
    assert data.get('ArtistId') == 1


def test_fetch_disabled(client):
    res = client.fetch('/album', json={})
    assert res.status_code == 405


def test_resources_list_disabled(client):
    res = client.get('/resources')
    assert res.status_code == 404


def test_meta_disabled(client):
    res = client.get('/artist/meta')
    assert res.status_code == 404


def test_export_disabled(client):
    res = client.get('/artist?_export=pippo')
    assert res.status_code == 200
    assert res.headers.get('Content-Type') == 'application/json'



import pytest

from . import create_app
from . import assert_export
from . import assert_pagination


@pytest.fixture
def app():
    return create_app()


@pytest.fixture
def client(app):
    _client = app.test_client()
    return _client


def test_fetch(client):
    res = client.fetch('/artist')
    assert res.status_code == 200

    res = client.fetch('/artist?_export=pippo')
    assert res.status_code == 200
    assert_export(res, 'pippo')

    res = client.fetch('/artist?_page=1&_limit=5')
    assert_pagination(res, 206, '1', '5')


def test_related(client):
    res = client.fetch(
        '/artist',
        json={"related": {"Album": ["*"]}}
    )
    assert res.status_code == 200

    data = res.get_json()
    assert data['ArtistList'][0].get('AlbumList') is not None


def test_filter(client):
    res = client.fetch(
        '/artist',
        json={
            "filters": [
                {
                    "model": "Artist",
                    "field": "ArtistId",
                    "op": "==",
                    "value": 1
                }
            ]
        }
    )
    assert res.status_code == 200

    artists = 'ArtistList'
    data = res.get_json()
    assert len(data[artists]) == 1
    assert data[artists][0].get('ArtistId') == 1


def test_validators(client):
    res = client.fetch(
        '/customer',
        json={
            "fields": "CustomerId",
            "related": {
                "Employee": "*",
                "Invoice": [1, 2, 3]
            },
            "filters": [
                {
                    "mode": "Invoice",
                    "fiel": "InvoiceDate",
                    "o": ">=",
                    "valu": "2010-04-01T00:00:00"
                }
            ],
            "sorting": [
                {
                    "mode": 1,
                    "fiel": "Total",
                    "directio": "asc"
                }
            ]
        }
    )
    assert res.status_code == 422

    data = res.get_json()['response']
    assert all(e in data for e in (
        "fields",
        "related",
        "sorting.0.field",
        "sorting.0.model",
        "sorting.0.direction",
        "filters.0.field",
        "filters.0.model",
        "filters.0.value"
    ))

    res = client.fetch(
        '/customer',
        json={
            "filters": 1,
            "related": 2
        }
    )
    assert res.status_code == 422

    data = res.get_json()['response']
    assert 'filters' in data
    assert 'related' in data

    res = client.fetch(
        '/artist',
        json={
            "filters": [
                {
                    "model": "Artist",
                    "field": "pluto",
                    "op": "==",
                    "value": 1
                }
            ]
        }
    )
    assert res.status_code == 400

    data = res.get_json()['response']
    assert 'invalid' in data and len(data['invalid']) == 1
    assert 'pluto' in data['invalid']

    res = client.fetch(
        '/customer',
        json={
            "fields": ["pippo"],
            "related": {
                "Employee": ["pluto"],
                "Invoice": ["paperino"]
            }
        }
    )
    assert res.status_code == 400

    data = res.get_json()['response']
    assert 'invalid' in data and len(data['invalid']) == 3
    assert all(e in data['invalid'] for e in (
        "pippo",
        "pluto",
        "paperino"
    ))



import pytest

from . import create_app

from .models import artists
from .models import albums


@pytest.fixture
def app():
    return create_app(
        m=[albums, artists],
        conf={'AUTOCRUD_CONDITIONAL_REQUEST_ENABLED': False}
    )


@pytest.fixture
def client(app):
    _client = app.test_client()
    return _client


def test_resource_crud(client):
    res = client.post(
        '/artists',
        json={'name': 'pippo'}
    )
    assert res.status_code == 201
    assert res.headers.get('Content-Type') == 'application/json'
    assert res.headers.get('ETag') is None

    data = res.get_json()
    id = data.get('id')

    assert res.headers.get('Location').endswith('/artists/{}'.format(id))

    res = client.post(
        '/artists',
        json={'name': 'pippo'},
        headers={'Accept': 'application/xml'}
    )
    assert res.status_code == 409
    assert res.headers.get('Content-Type') == 'application/xml; charset=utf-8'

    res = client.get('/artists/{}'.format(id))
    assert res.status_code == 200
    assert res.headers.get('ETag') is None
    assert res.headers.get('Content-Type') == 'application/json'
    assert res.headers.get('Link') == "</artists/{id}>; rel=self, </artists/{id}/myalbum>; rel=related".format(id=id)

    data = res.get_json()
    returned_id = data.get('id')
    assert returned_id == id

    res = client.put(
        '/artists/{}'.format(id),
        json={'name': 'pippo2'}
    )
    assert res.status_code == 200
    assert res.headers.get('ETag') is None

    res = client.patch(
        '/artists/{}'.format(id),
        json={'name': 'pippo3'}
    )
    assert res.status_code == 200
    assert res.headers.get('ETag') is None

    res = client.delete('/artists/{}'.format(id))
    assert res.status_code == 204


def test_resource_meta(client):
    res = client.get('/artists/meta')
    assert res.status_code == 200
    assert res.headers.get('Content-Type') == 'application/json'

    data = res.get_json()
    assert all(i in data.keys() for i in (
        'name',
        'description',
        'fields',
        'methods',
        'url'
    ))
    assert data.get('description') == 'artists'


def test_hateoas(client):
    res = client.get('/artists/1')
    assert res.status_code == 200

    data = res.get_json()
    assert '_links' in data.keys()
    assert data['_links'].get('self') == '/artists/1'


def test_extended(client):
    res = client.get('/myalbum/5?_extended')
    assert res.status_code == 200

    data = res.get_json()
    assert data['id'] == 5
    assert isinstance(data["artists"], dict)


def test_extended_list(client):
    res = client.get('/myalbum?_extended')
    assert res.status_code == 200

    data = res.get_json()['albumsList'][0]
    assert isinstance(data["artists"], dict)
    assert data.get('_links') is not None


def test_fields(client):
    res = client.get('/artists')
    assert res.status_code == 200

    data = res.get_json()['artistsList']
    assert len(data[0].keys()) == 3

    res = client.get('/artists?_fields=id')
    assert res.status_code == 200

    data = res.get_json()['artistsList']
    assert len(data[0].keys()) == 2
    assert all(e in data[0].keys() for e in (
        "id",
        "_links"
    ))


def test_sorting(client):
    res = client.get('/artists?_sort=id')
    assert res.status_code == 200

    data = res.get_json()['artistsList']
    first_id = data[0].get('id')

    res = client.get('/artists?_sort=-id')
    assert res.status_code == 200

    data = res.get_json()['artistsList']
    last_id = data[0].get('id')
    assert last_id != first_id


def test_range(client):
    res = client.get('/artists?id=(1;3)')
    assert res.status_code == 200

    data = res.get_json()['artistsList']
    assert len(data) == 3
    assert data[0].get('id') == 1
    assert data[1].get('id') == 2
    assert data[2].get('id') == 3


def test_null(client):
    res = client.get('/artists?id=null')
    assert res.status_code == 200

    data = res.get_json()['artistsList']
    assert len(data) == 0


def test_related(client):
    res = client.fetch(
        '/myalbum',
        json={"related": {"artists": ["*"]}}
    )
    assert res.status_code == 200

    data = res.get_json()['albumsList']
    assert data[0].get('artists') is not None


def test_filter(client):
    res = client.fetch(
        '/artists',
        json={
            "filters": [
                {
                    "model": "artists",
                    "field": "id",
                    "op": "==",
                    "value": 1
                }
            ]
        }
    )
    assert res.status_code == 200

    data = res.get_json()['artistsList']
    assert len(data) == 1
    assert data[0].get('id') == 1


def test_subresource(client):
    res = client.get('/artists/1/myalbum?_extended')
    assert res.status_code == 200

    data = res.get_json()['albumsList'][0]
    assert data['artist_id'] == 1
    assert all(e in data.keys() for e in (
        "id",
        "artists"
    ))


def test_hidden_field(client):
    res = client.get('/myalbum/5')
    assert res.status_code == 200

    data = res.get_json()
    assert data['id'] == 5
    assert len(data.keys()) == 3

    res = client.get('/myalbum/meta')
    assert res.status_code == 200

    data = res.get_json()
    assert len(data['fields']) == 2



import flask
import colander

from flask import request
from flask import current_app as cap
from flask.views import MethodView

from werkzeug.http import generate_etag
from werkzeug.exceptions import NotImplemented
from werkzeug.exceptions import MethodNotAllowed

import sqlalchemy_filters as sqlaf
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.attributes import ScalarObjectAttributeImpl

from flask_response_builder.dictutils import to_flatten

from .qs2sqla import Qs2Sqla
from .config import HttpStatus as status
from .validators import FetchPayloadSchema


class Service(MethodView):
    _db = None
    _model = None
    _response = None
    syntax = None
    arguments = None

    def dispatch_request(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        """
        methods = self._model.__methods__
        if request.method not in methods:
            raise MethodNotAllowed(valid_methods=list(methods))

        controller = getattr(self, request.method.lower(), None)
        if controller is None:
            raise NotImplemented()

        return controller(*args, **kwargs)

    def delete(self, resource_id):
        """

        :param resource_id:
        :return:
        """
        @self._response.no_content
        def _delete():
            mime_type, _ = self._response.get_mimetype_accept()

            model = self._model
            session = self._db.session()
            resource = model.query.get(resource_id)

            if not resource:
                flask.abort(status.NOT_FOUND)

            self._check_etag(resource)
            session.delete(resource)
            session.commit()
        return _delete()

    def post(self):
        """

        :return:
        """
        model = self._model
        _, builder = self._response.get_mimetype_accept()
        data = self._validate_new_data()

        resource = model.query.filter_by(**data).first()

        try:
            if resource:
                raise IntegrityError(statement=None, params=None, orig=None)

            resource = model(**data)
            res = self._add_resource(resource)
        except IntegrityError:
            self._db.session().rollback()
            flask.abort(status.CONFLICT)
            return  # only to prevent warning

        return self._response_with_etag(
            builder, (res, status.CREATED, self._location_header(resource)), res
        )

    def put(self, resource_id):
        """

        :param resource_id:
        :return:
        """
        model = self._model
        _, builder = self._response.get_mimetype_accept()
        data = self._validate_new_data()

        resource = model.query.get(resource_id)
        if resource:
            self._check_etag(resource)
            res = self._merge_resource(resource, data)
            return self._response_with_etag(
                builder, (res, self._link_header(resource)), res
            )

        resource = model(**{model.primary_key_field(): resource_id, **data})
        res = self._add_resource(resource)
        return self._response_with_etag(
            builder, (res, self._location_header(resource), status.CREATED), res
        )

    def patch(self, resource_id):
        """

        :param resource_id:
        :return:
        """
        model = self._model
        _, builder = self._response.get_mimetype_accept()
        data = request.get_json() or {}
        _, unknown = model.validate(data)

        if unknown:
            flask.abort(status.UNPROCESSABLE_ENTITY, response=dict(unknown=unknown))

        resource = model.query.get(resource_id)
        if not resource:
            flask.abort(status.NOT_FOUND)

        self._check_etag(resource)
        res = self._merge_resource(resource, data)
        return self._response_with_etag(
            builder, (res, self._link_header(resource)), res
        )

    def get(self, resource_id=None, subresource=None):
        """

        :param resource_id:
        :param subresource:
        :return:
        """
        related = {}
        model = self._model
        _, builder = self._response.get_mimetype_accept()

        filter_by_id = [
            Qs2Sqla(model, self.syntax, self.arguments).get_filter(
                model.primary_key_field(), str(resource_id)
            )
        ]

        if request.path.endswith(cap.config['AUTOCRUD_METADATA_URL']):
            return self._response.build_response(builder, model.description())

        if subresource is not None:
            model = model.submodel_from_url("/" + subresource)
            if not model:
                flask.abort(status.NOT_FOUND)

        qsqla = Qs2Sqla(model, self.syntax, self.arguments)
        if qsqla.arguments.scalar.extended in request.args:
            for k, v in model.related().items():
                if isinstance(v['instance'].impl, ScalarObjectAttributeImpl):
                    related.update({k: "*"})

        if resource_id is not None:
            query, _ = qsqla.dict2sqla(dict(filters=filter_by_id, related=related))

            if subresource is None:
                resource = query.one_or_none()
                if not resource:
                    flask.abort(status.NOT_FOUND)

                res = resource.to_dict(links=True)
                self._check_etag(res)

                return self._response_with_etag(
                    builder, (res, self._link_header(resource)), res
                )

        if cap.config['AUTOCRUD_QUERY_STRING_FILTERS_ENABLED'] is True:
            data, error = qsqla.parse(request.args)
        else:
            data, error = {}, []

        if resource_id is not None:
            if data.get('filters') is None:
                data['filters'] = filter_by_id
            else:
                data['filters'] += filter_by_id

        return self._build_response_list(model, builder, {**data, 'related': related}, error)

    def fetch(self):
        """

        :return:
        """
        _, builder = self._response.get_mimetype_accept()

        try:
            schema = FetchPayloadSchema()
            data = schema.deserialize(request.get_json() or {})
        except colander.Invalid as exc:
            flask.abort(status.UNPROCESSABLE_ENTITY, response=exc.asdict())
            return  # only to prevent warning

        return self._build_response_list(self._model, builder, data)

    def _build_response_list(self, model, builder, data, error=None):
        """

        :param model:
        :param builder:
        :param data:
        :param error:
        :return:
        """
        qsqla = Qs2Sqla(model, self.syntax, self.arguments)
        invalid = error or []

        page, limit, error = qsqla.get_pagination(
            request.args,
            cap.config['AUTOCRUD_MAX_QUERY_LIMIT']
        )
        invalid += error

        query, error = qsqla.dict2sqla(data)
        invalid += error

        if len(invalid) > 0:
            flask.abort(status.BAD_REQUEST, response=dict(invalid=invalid))

        query, pagination = sqlaf.apply_pagination(query, page, limit)
        result = query.all()

        response = []
        links_enabled = cap.config['AUTOCRUD_EXPORT_ENABLED'] is False \
            or qsqla.arguments.scalar.export not in request.args

        for r in result:
            if qsqla.arguments.scalar.as_table in request.args:
                response += to_flatten(r, to_dict=model.to_dict)
            else:
                response.append(r.to_dict(links=links_enabled))

        if cap.config['AUTOCRUD_EXPORT_ENABLED'] is True:
            if qsqla.arguments.scalar.export in request.args:
                filename = request.args.get(qsqla.arguments.scalar.export) or "{}{}{}".format(
                    self._model.__name__,
                    "_{}".format(page) if page else "",
                    "_{}".format(limit) if limit else ""
                )
                return self._response.csv(response, filename=filename)

        response = {
            model.__name__ + model.collection_suffix: response,
            '_meta': self._pagination_meta(pagination)
        }

        etag = self._compute_etag(response)
        self._check_etag(etag)

        return self._response_with_etag(
            builder, (response, *self._pagination_headers(pagination)), etag
        )

    def _response_with_etag(self, builder, data, etag):
        """

        :param builder:
        :param data:
        :param etag:
        :return:
        """
        response = self._response.build_response(builder, data)

        if cap.config['AUTOCRUD_CONDITIONAL_REQUEST_ENABLED'] is True:
            response.set_etag(etag if isinstance(etag, str) else self._compute_etag(etag))

        return response

    def _validate_new_data(self):
        """

        :return:
        """
        model = self._model
        data = request.get_json()

        if not data:
            flask.abort(status.BAD_REQUEST)

        missing, unknown = model.validate(data)
        if unknown or missing:
            flask.abort(status.UNPROCESSABLE_ENTITY, response=dict(
                unknown=unknown or [], missing=missing or []
            ))

        return data

    @classmethod
    def _add_resource(cls, resource):
        """

        :param resource:
        :return:
        """
        session = cls._db.session
        session.add(resource)
        session.flush()
        res = resource.to_dict(links=True)
        session.commit()
        return res

    @classmethod
    def _merge_resource(cls, resource, data):
        """

        :param resource:
        :param data:
        :return:
        """
        session = cls._db.session
        session.merge(resource)
        resource.update(data)
        session.flush()
        res = resource.to_dict(links=True)
        session.commit()
        return res

    @classmethod
    def _pagination_meta(cls, pagination):
        """

        :param pagination:
        :return:
        """
        args = Qs2Sqla(cls._model).arguments.scalar

        page_number = pagination.page_number
        num_pages = pagination.num_pages
        page_size = pagination.page_size

        def get_link(p):
            return "{}?{}={}&{}={}".format(request.path, args.page, p, args.limit, page_size)

        return dict(
            first=get_link(1) if page_number > 1 else None,
            last=get_link(num_pages) if page_number != num_pages else None,
            next=get_link(page_number + 1) if page_number != num_pages else None,
            prev=get_link(page_number - 1) if page_number != 1 else None
        )

    @classmethod
    def _pagination_headers(cls, pagination):
        """

        :param pagination:
        :return:
        """
        code = status.SUCCESS

        if not pagination:
            return {}, code

        total_results = pagination.total_results
        page_number = pagination.page_number
        num_pages = pagination.num_pages
        page_size = pagination.page_size

        if num_pages > 1 and total_results > page_size:
            code = status.PARTIAL_CONTENT

        if page_number == num_pages:
            code = status.SUCCESS

        return {
            'Pagination-Count': total_results,
            'Pagination-Page': page_number,
            'Pagination-Num-Pages': num_pages,
            'Pagination-Page-Size': page_size,
            **(cls._link_header(None, **cls._pagination_meta(pagination)) or {})
        }, code

    @staticmethod
    def _link_header(resource=None, **kwargs):
        """

        :param resource:
        :return:
        """
        links = []

        if resource is not None:
            res_links = resource.links()
            links.append('<{}>; rel=self'.format(res_links.get('self')))

            for k, l in res_links.items():
                if k != 'self':
                    links.append('<{}>; rel=related'.format(l))

        for k, l in kwargs.items():
            if l is not None:
                links.append('<{}>; rel={}'.format(l, k))

        return dict(Link=", ".join(links)) if len(links) else {}

    @staticmethod
    def _location_header(resource):
        """

        :param resource:
        :return:
        """
        location = resource.links()
        return dict(Location=location.get('self'))

    @classmethod
    def _compute_etag(cls, data):
        """

        :param data:
        :return:
        """
        if cap.config['AUTOCRUD_CONDITIONAL_REQUEST_ENABLED'] is True:
            if not isinstance(data, str):
                data = str(data if isinstance(data, (dict, list)) else data.to_dict(True))
            return generate_etag(data.encode('utf-8'))
        return ""

    @classmethod
    def _check_etag(cls, data):
        """

        :param data:
        :return:
        """
        if cap.config['AUTOCRUD_CONDITIONAL_REQUEST_ENABLED'] is True:
            match = request.if_match
            none_match = request.if_none_match
            etag = data if isinstance(data, str) else cls._compute_etag(data)

            if request.method in ('GET', 'FETCH'):
                if none_match and etag in none_match:
                    flask.abort(flask.Response(status=status.NOT_MODIFIED))
            elif request.method in ('PUT', 'PATCH', 'DELETE'):
                if not match:
                    flask.abort(status.PRECONDITION_REQUIRED)
                elif etag not in match:
                    flask.abort(status.PRECONDITION_FAILED, response={'invalid': etag})



"""
Flask-AutoCRUD
-------------

Automatically generated a RESTful API services for CRUD operation and queries on database
"""
import sys
import pytest

from setuptools import setup
from setuptools import find_packages
from setuptools.command.test import test

from flask_autocrud import __version__
from flask_autocrud import __author_info__


with open("README.rst") as fh:
    long_description = fh.read()


class PyTest(test):
    def finalize_options(self):
        """

        """
        test.finalize_options(self)

    def run_tests(self):
        """

        """
        sys.exit(pytest.main(['tests']))


setup(
    name='Flask-AutoCRUD',
    version=__version__,
    url='https://github.com/cs91chris/flask_autocrud/',
    license='MIT',
    author=__author_info__['name'],
    author_email=__author_info__['email'],
    description='Automatically generated a RESTful API services for CRUD operation and queries on database',
    long_description=long_description,
    packages=find_packages(),
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    entry_points={
        'console_scripts': [
            'autocrud = flask_autocrud.scripts.run:main',
        ],
    },
    tests_require=[
        'pytest==4.5.0',
        'pytest-cov==2.7.1'
    ],
    install_requires=[
        'Flask==1.1.*',
        'Flask-SQLAlchemy==2.4.*',
        'Flask-ResponseBuilder==2.*',
        'Flask-ErrorsHandler==2.*',
        'sqlalchemy-filters==0.10.*',
        'colander==1.7.*',
        'PyYAML==5.*'
    ],
    cmdclass={'test': PyTest},
    test_suite='tests',
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



from collections import namedtuple

from werkzeug.routing import FloatConverter
from werkzeug.routing import UnicodeConverter


class HttpStatus:
    SUCCESS = 200
    CREATED = 201
    NO_CONTENT = 204
    PARTIAL_CONTENT = 206
    NOT_MODIFIED = 304
    BAD_REQUEST = 400
    NOT_FOUND = 404
    CONFLICT = 409
    PRECONDITION_FAILED = 412
    UNPROCESSABLE_ENTITY = 422
    PRECONDITION_REQUIRED = 428
    INTERNAL_SERVER_ERROR = 500


ALLOWED_METHODS = {
    'GET',
    'POST',
    'PUT',
    'PATCH',
    'DELETE',
    'FETCH',
    'HEAD',
    'OPTIONS'
}


def set_default_config(app):
    """

    :param app:
    """
    app.config.setdefault('AUTOCRUD_METADATA_ENABLED', True)
    app.config.setdefault('AUTOCRUD_METADATA_URL', '/meta')
    app.config.setdefault('AUTOCRUD_READ_ONLY', False)
    app.config.setdefault('AUTOCRUD_BASE_URL', '')
    app.config.setdefault('AUTOCRUD_RESOURCES_URL', '/resources')
    app.config.setdefault('AUTOCRUD_SUBDOMAIN', None)
    app.config.setdefault('AUTOCRUD_DATABASE_SCHEMA', None)
    app.config.setdefault('AUTOCRUD_RESOURCES_URL_ENABLED', True)
    app.config.setdefault('AUTOCRUD_MAX_QUERY_LIMIT', 1000)
    app.config.setdefault('AUTOCRUD_FETCH_ENABLED', True)
    app.config.setdefault('AUTOCRUD_EXPORT_ENABLED', True)
    app.config.setdefault('AUTOCRUD_QUERY_STRING_FILTERS_ENABLED', True)
    app.config.setdefault('AUTOCRUD_CONDITIONAL_REQUEST_ENABLED', True)

    app.url_map.converters.update({
        'str': UnicodeConverter,
        'decimal': FloatConverter,
    })


Syntax = namedtuple(
    'Syntax',
    (
        'ESCAPE',
        'ALL',
        'SEP',
        'NOT',
        'LIKE',
        'REVERSE',
        'LT',
        'GT',
        'LTE',
        'GTE',
        'NULL',
        'RNS',
        'RNE',
        'NOT_RNS',
        'NOT_RNE',
        'NOT_NULL',
        'NOT_LIKE',
    )
)

scalarFields = namedtuple(
    'scalarFields',
    (
        'page',
        'limit',
        'export',
        'extended',
        'as_table',
    )
)

vectorFields = namedtuple(
    'vector',
    (
        'sort',
        'fields',
    )
)


class Fields:
    scalar = scalarFields(
        page='_page',
        limit='_limit',
        export='_export',
        extended='_extended',
        as_table='_as_table',
    )

    vector = vectorFields(
        sort='_sort',
        fields='_fields',
    )


default_syntax = Syntax(
    ESCAPE='\\',
    ALL='*',
    SEP=';',
    NOT='!',
    REVERSE='-',
    LIKE='%',
    NOT_LIKE='!%',
    LT='__lt__',
    GT='__gt__',
    LTE='__lte__',
    GTE='__gte__',
    NULL='null',
    NOT_NULL='!null',
    RNS='(',
    RNE=')',
    NOT_RNS='!(',
    NOT_RNE=')',
)

default_arguments = Fields



import yaml
import argparse

from yaml.error import YAMLError

from flask import Flask

from flask_autocrud import AutoCrud
from flask_sqlalchemy import SQLAlchemy
from flask_errors_handler import ErrorHandler

try:
    from .gunicorn import StandaloneApplication
except (ModuleNotFoundError, ImportError):
    try:
        from .waitress import StandaloneApplication
    except (ModuleNotFoundError, ImportError):
        from .default import StandaloneApplication


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-D', '--database', default=None)
    parser.add_argument('-d', '--debug', action='store_true')
    parser.add_argument('-b', '--bind', default='0.0.0.0:5000')
    parser.add_argument('-w', '--workers', default=1)
    parser.add_argument('-c', '--config', default=None)
    args = parser.parse_args()

    if args.config is not None:
        try:
            with open(args.config) as f:
                config = yaml.safe_load(f)
        except (FileNotFoundError, YAMLError) as exc:
            import sys
            print(exc, file=sys.stderr)
            sys.exit(1)
    else:
        config = {
            'app': {
                'DEBUG': args.debug,
                'SQLALCHEMY_DATABASE_URI': args.database,
                'SQLALCHEMY_TRACK_MODIFICATIONS': False
            },
            'wsgi': {
                'bind': args.bind,
                'workers': args.workers,
                'debug': args.debug
            }
        }

    app = Flask(__name__)
    app.config.update(config.get('app', {}))

    error = ErrorHandler(app)
    error.api_register(app)
    AutoCrud(app, SQLAlchemy(app), error=error)

    StandaloneApplication(app, config.get('wsgi', {})).run()


if __name__ == '__main__':
    main()



# noinspection PyUnresolvedReferences
from gunicorn.six import iteritems

# noinspection PyUnresolvedReferences
from gunicorn.app.base import BaseApplication


class StandaloneApplication(BaseApplication):
    def __init__(self, app, options=None):
        """

        :param app:
        :param options:
        """
        self.application = app
        self.options = options or {}
        super(StandaloneApplication, self).__init__()

    def init(self, parser, opts, args):
        """

        :param parser:
        :param opts:
        :param args:
        """
        pass

    def load_config(self):
        """

        """
        for key, value in iteritems(dict(
            [
                (k, v) for k, v in iteritems(self.options)
                if k in self.cfg.settings and v is not None
            ]
        )):
            self.cfg.set(key.lower(), value)

    def load(self):
        """

        :return:
        """
        return self.application



from flask import Flask

from flask_admin import Admin
from flask_admin.contrib.sqla import ModelView

from flask_autocrud import AutoCrud
from flask_sqlalchemy import SQLAlchemy


class CustomAdminView(ModelView):
    can_export = True
    details_modal = True
    column_display_pk = True
    can_set_page_size = True
    can_view_details = True


def main():
    app = Flask(__name__)
    app.config['SECRET_KEY'] = 'more_difficult_string'
    app.config['FLASK_ADMIN_SWATCH'] = 'cosmo'
    app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite+pysqlite:///examples/db.sqlite3'
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False

    admin = Admin(app)
    db = SQLAlchemy(app)
    autocrud = AutoCrud(app, db)

    for k, m in autocrud.models.items():
        setattr(CustomAdminView, 'column_searchable_list', m.searchable())
        admin.add_view(CustomAdminView(m, db.session))

    app.run(debug=True)


if __name__ == '__main__':
    main()



import pytest

from . import create_app
from . import assert_pagination


@pytest.fixture
def app():
    return create_app()


@pytest.fixture
def client(app):
    _client = app.test_client()
    return _client


def test_pagination(client):
    res = client.get('/artist?_page=1&_limit=5')
    assert_pagination(res, 206, '1', '5')


def test_extended(client):
    res = client.get('/track/5?_extended')
    assert res.status_code == 200

    data = res.get_json()
    assert data['TrackId'] == 5
    assert all(isinstance(data[e], dict) for e in (
        "Album",
        "Genre",
        "MediaType"
    ))


def test_extended_list(client):
    res = client.get('/track?_extended')
    assert res.status_code == 206

    tracks = 'TrackList'
    data = res.get_json()
    assert data[tracks][0].get('_links') is not None
    assert all(isinstance(data[tracks][0][e], dict) for e in (
        "Album",
        "Genre",
        "MediaType"
    ))


def test_fields(client):
    res = client.get('/artist')
    assert res.status_code == 200

    artists = 'ArtistList'
    data = res.get_json()
    assert len(data[artists][0].keys()) == 3

    res = client.get('/artist?_fields=ArtistId')
    assert res.status_code == 200

    data = res.get_json()
    assert len(data[artists][0].keys()) == 2
    assert all(e in data[artists][0].keys() for e in (
        "ArtistId",
        "_links"
    ))


def test_sorting(client):
    res = client.get('/artist?_sort=ArtistId')
    assert res.status_code == 200

    data = res.get_json()
    first_id = data['ArtistList'][0].get('ArtistId')

    res = client.get('/artist?_sort=-ArtistId')
    assert res.status_code == 200

    data = res.get_json()
    last_id = data['ArtistList'][0].get('ArtistId')
    assert last_id != first_id


def test_comparator(client):
    res = client.get('/invoice?Total=__gt__25')
    assert res.status_code == 200

    invoices = 'InvoiceList'
    data = res.get_json()
    assert len(data[invoices]) == 1

    res = client.get('/invoice?Total=__gte__25')
    assert res.status_code == 200

    invoices = 'InvoiceList'
    data = res.get_json()
    assert len(data[invoices]) == 1

    res = client.get('/artist?ArtistId=__lte__2')
    assert res.status_code == 200

    artists = 'ArtistList'
    data = res.get_json()
    assert len(data[artists]) == 2

    res = client.get('/artist?ArtistId=__lt__2')
    assert res.status_code == 200

    artists = 'ArtistList'
    data = res.get_json()
    assert len(data[artists]) == 1
    assert data[artists][0].get('ArtistId') == 1

    links = data[artists][0].get('_links')
    assert links is not None
    assert links.get('Album') == "/artist/1/album"
    assert links.get('self') == "/artist/1"


def test_like(client):
    res = client.get('/album?Title=%...%')
    assert res.status_code == 200

    albums = 'AlbumList'
    data = res.get_json()[albums][0]
    assert data['Title'].startswith('...')

    res = client.get('/album?Title=%%z')
    assert res.status_code == 200

    data = res.get_json()[albums][0]
    assert data['Title'].endswith('z')

    res = client.get('/artist?_sort=Name&Name=!%A%')
    assert res.status_code == 200

    albums = 'ArtistList'
    data = res.get_json()[albums][0]
    assert data['Name'].startswith('B')


def test_range(client):
    res = client.get('/artist?ArtistId=(1;3)')
    assert res.status_code == 200

    artists = 'ArtistList'
    data = res.get_json()
    assert len(data[artists]) == 3
    assert data[artists][0].get('ArtistId') == 1
    assert data[artists][1].get('ArtistId') == 2
    assert data[artists][2].get('ArtistId') == 3

    res = client.get('/invoice?InvoiceDate=!(2008-01-01;2013-12-20 00:00:00)')
    assert res.status_code == 200
    data = res.get_json()
    assert len(data['InvoiceList']) == 1


def test_null(client):
    res = client.get('/artist?ArtistId=null')
    assert res.status_code == 200

    data = res.get_json()
    assert len(data['ArtistList']) == 0

    res = client.get('/track?AlbumId=!null')
    assert res.status_code == 206

    data = res.get_json()
    meta = data.get('_meta')
    assert meta['first'] is None
    assert meta['prev'] is None
    assert meta['next'] == "/track?_page=2&_limit=1000"
    assert meta['last'] == "/track?_page=4&_limit=1000"


def test_query_string_invalid(client):
    res = client.get('/invoice?_fields=pippo&_sort=pluto&paperino=1')
    assert res.status_code == 400

    data = res.get_json()['response']
    assert 'invalid' in data and len(data['invalid']) == 3
    assert all(e in data['invalid'] for e in (
        'pippo',
        'pluto',
        'paperino'
    ))


def test_check_etag_list(client):
    res = client.get('/album/5/track?_extended')

    etag = res.headers.get('ETag')
    assert res.status_code == 200
    assert etag is not None

    res = client.get('/album/5/track?_extended', headers={'If-None-Match': etag})
    assert res.status_code == 304



from sqlalchemy import ForeignKey
from sqlalchemy.orm import relationship

from flask_autocrud import Model
from flask_sqlalchemy import SQLAlchemy


db = SQLAlchemy()


class artists(db.Model, Model):
    __tablename__ = "Artist"
    __description__ = "artists"
    id = db.Column('ArtistId', db.Integer, primary_key=True, comment='primarykey')
    name = db.Column('Name', db.String(80), unique=True, nullable=False)


class albums(db.Model, Model):
    __tablename__ = "Album"
    __hidden__ = "title"
    __url__ = '/myalbum'
    id = db.Column('AlbumId', db.Integer, primary_key=True)
    title = db.Column('Title', db.String(80), unique=True, nullable=False)
    artist_id = db.Column('ArtistId', db.Integer, ForeignKey("Artist.ArtistId"), nullable=False)
    artists = relationship(artists, backref="albums")



from sqlalchemy.exc import ArgumentError
from sqlalchemy.orm import contains_eager

import sqlalchemy_filters as sqlaf
from sqlalchemy_filters import exceptions

from .config import Syntax
from .config import Fields
from .config import default_syntax
from .config import default_arguments


class Qs2Sqla:
    def __init__(self, model, syntax=None, arguments=None):
        """

        :param model:
        :param syntax:
        :param arguments:
        """
        self._model = model
        self._syntax = syntax or default_syntax
        self._arguments = arguments or default_arguments

        if self._syntax and not isinstance(self._syntax, Syntax):
            raise AttributeError(
                "'{}' must be an instance of {}".format(
                    self._syntax.__name__, Syntax.__name__
                )
            )

        if self._arguments and not issubclass(self._arguments, Fields):
            raise AttributeError(
                "'{}' must be a subclass of {}".format(
                    self._arguments.__name__, Fields.__name__
                )
            )

    @property
    def syntax(self):
        """

        :return:
        """
        return self._syntax

    @property
    def arguments(self):
        """

        :return:
        """
        return self._arguments

    def clear_empty(self, l, sep=None):
        """

        :param l:
        :param sep:
        :return:
        """
        return [i for i in l.split(sep or self._syntax.SEP) if i != ""]

    def clear_escape(self, i, escape=None):
        """

        :param i:
        :param escape:
        :return:
        """
        esc = escape or self._syntax.ESCAPE
        return i[len(esc):] if i.startswith(esc) else i

    def get_pagination(self, conf, max_limit):
        """

        :param conf:
        :param max_limit:
        :return:
        """
        def valid_number(num):
            if num is None:
                return None

            try:
                num = int(num)
                return num if num > 0 else False
            except ValueError:
                return False

        invalid = []
        max_limit = valid_number(max_limit)
        page = valid_number(conf.get(self._arguments.scalar.page))
        limit = valid_number(conf.get(self._arguments.scalar.limit))

        if page is False:
            invalid.append(self._arguments.scalar.page)
        if limit is False:
            invalid.append(self._arguments.scalar.limit)
        if max_limit is False:
            invalid.append('invalid max_limit: {}'.format(max_limit))

        if max_limit > 0:
            page = 1 if not page else page
            if not limit or limit > max_limit:
                limit = max_limit

        return page, limit, invalid

    def get_filter(self, f, v):
        """

        :param f:
        :param v:
        :return:
        """
        def to_dict(op, value):
            """

            :param op:
            :param value:
            :return:
            """
            return dict(model=self._model.__name__, field=f, op=op, value=value)

        if v.startswith(self._syntax.GT):
            return to_dict('>', self.clear_escape(v, escape=self._syntax.GT))
        if v.startswith(self._syntax.LT):
            return to_dict('<', self.clear_escape(v, escape=self._syntax.LT))
        if v.startswith(self._syntax.GTE):
            return to_dict('>=', self.clear_escape(v, escape=self._syntax.GTE))
        if v.startswith(self._syntax.LTE):
            return to_dict('<=', self.clear_escape(v, escape=self._syntax.LTE))
        if v.startswith(self._syntax.NOT_LIKE):
            return {'not': [to_dict('like', self.clear_escape(v, escape=self._syntax.NOT_LIKE))]}
        if v.startswith(self._syntax.LIKE):
            return to_dict('like', self.clear_escape(v, escape=self._syntax.LIKE))

        if v.startswith(self._syntax.RNS) and v.endswith(self._syntax.RNE):
            down, up = v[len(self._syntax.RNS):-len(self._syntax.RNE)].split(self._syntax.SEP, 1)
            return {'and': [to_dict('>=', down), to_dict('<=', up)]}

        if v.startswith(self._syntax.NOT_RNS) and v.endswith(self._syntax.NOT_RNE):
            down, up = v[len(self._syntax.NOT_RNS):-len(self._syntax.NOT_RNE)].split(self._syntax.SEP, 1)
            return {'or': [to_dict('<', down), to_dict('>', up)]}

        item = self.clear_empty(v)

        if item[0].startswith(self._syntax.NOT):
            item[0] = self.clear_escape(item[0], escape=self._syntax.NOT)
            return to_dict('not_in', item)

        item[0] = self.clear_escape(item[0])
        return to_dict('in', item)

    def parse(self, args):
        """

        :param args:
        :return:
        """
        invalid = []
        resp = dict(fields=[], filters=[], sorting=[])

        for k, v in args.items():
            if k in self._arguments.scalar:
                continue

            if k == self._arguments.vector.sort:
                for item in self.clear_empty(v):
                    d = 'desc' if item.startswith(self._syntax.REVERSE) else 'asc'
                    item = self.clear_escape(item, escape=self._syntax.REVERSE)

                    if item in self._model.columns().keys():
                        resp['sorting'].append(dict(field=item, direction=d))
                    else:
                        invalid.append(item)

            elif k == self._arguments.vector.fields:
                for item in self.clear_empty(v):
                    if item in self._model.columns().keys():
                        resp['fields'].append(item)
                    else:
                        invalid.append(item)

            elif k in self._model.columns().keys():
                resp['filters'].append({
                    'or': [self.get_filter(k, item) for item in args.getlist(k)]
                })
            else:
                invalid.append(k)
        return resp, invalid

    def dict2sqla(self, data):
        """

        :param data:
        :return:
        """
        invalid = []
        query = self._model.query
        fields = data.get('fields') or list(self._model.columns().keys())
        related = data.get('related') or {}
        filters = data.get('filters') or []
        sort = data.get('sorting') or []

        for k in fields:
            if k not in self._model.columns().keys():
                invalid.append(k)

        if len(invalid) == 0 and len(fields) > 0:
            try:
                query = sqlaf.apply_loads(query, fields)
            except exceptions.BadLoadFormat:
                invalid.append(fields)

        for k in related.keys():
            instance, columns = self._model.related(k)
            if instance is not None:
                _columns = related.get(k)
                try:
                    if len(_columns) > 0 and _columns[0] != self._syntax.ALL:
                        _invalid = list(set(related.get(k)) - set(columns))
                        if len(_invalid) > 0:
                            _columns = _invalid
                            raise ArgumentError
                    else:
                        _columns = columns

                    query = query.join(instance, aliased=False)
                    query = query.options(contains_eager(instance).load_only(*_columns))
                except ArgumentError:
                    invalid += _columns
            else:
                invalid.append(k)

        def apply(stm, flt, action):
            """

            :param stm:
            :param flt:
            :param action:
            :return:
            """
            resource = flt.get('model')
            try:
                if resource:
                    _, cols = self._model.related(resource)
                    if cols and cols.get(flt.get('field')) is None:
                        raise exceptions.FieldNotFound

                return action(stm, flt)
            except exceptions.BadSpec:
                invalid.append(resource)
            except exceptions.FieldNotFound:
                invalid.append(flt.get('field'))
            except exceptions.BadFilterFormat:
                invalid.append(flt.get('op'))
            except exceptions.BadSortFormat:
                invalid.append(flt.get('direction'))

        for f in filters:
            try:
                query = apply(query, f, sqlaf.apply_filters)
            except AttributeError:
                invalid.append(f)

        for s in sort:
            try:
                query = apply(query, s, sqlaf.apply_sort)
            except AttributeError:
                invalid.append(s)

        return query, invalid
